#
# This file is shipped by FreedomBox. DO NOT EDIT.
#
# This is to make Pagekite use certficates shipped by Debian. Otherwise by
# default, it uses internally shipped certificates that may be outdated. See:
# https://pagekite.wordpress.com/2020/05/30/tls-certificate-validation-issues/
#
# If you wish to override this setting, create another file starting with a
# number higher than 90.
#
ca_certs = /etc/ssl/certs/ca-certificates.crt
