{% extends "connections_edit.html" %}
{% comment %}
# SPDX-License-Identifier: AGPL-3.0-or-later
{% endcomment %}

{% load bootstrap %}
{% load i18n %}

{% block content %}

  <h3>{{ title }}</h3>

  <form class="form" method="post">
    {% csrf_token %}

    {{ form|bootstrap }}

    <input type="submit" class="btn btn-primary"
           value="{% trans "Create Connection" %}"/>
  </form>

{% endblock %}
