{% extends "base.html" %}
{% comment %}
# SPDX-License-Identifier: AGPL-3.0-or-later
{% endcomment %}

{% load bootstrap %}
{% load i18n %}

{% block content %}
  <h2>{{ title }}</h2>

  <p>{% trans "Delete this archive permanently?" %}</p>

  <div class="table-responsive">
    <table class="table">
      <thead>
        <th>{% trans "Name" %}</th>
        <th>{% trans "Time" %}</th>
      </thead>
      <tbody>
        <tr>
          <td>{{ archive.name }}</td>
          <td>{{ archive.time }}</td>
        </tr>
      </tbody>
    </table>
  </div>

  <p>
    <form class="form form-delete" method="post">
      {% csrf_token %}

      <input type="submit" class="btn btn-danger"
             value="{% blocktrans trimmed with name=archive.name %}
                      Delete Archive {{ name }}
                    {% endblocktrans %}"/>
    </form>
  </p>

{% endblock %}
