{% extends "app.html" %}
{% comment %}
# SPDX-License-Identifier: AGPL-3.0-or-later
{% endcomment %}

{% load i18n %}
{% load static %}

{% block status %}
  {{ block.super }}

  <h3>{% trans "Status" %}</h3>

  <p>
    {% blocktrans trimmed %}
      The Matrix server domain is set to <em>{{ domain_name }}</em>. User IDs
      will look like <em>@username:{{ domain_name }}</em>. Changing the domain
      name after the initial setup is currently not supported.
    {% endblocktrans %}
  </p>
  <p>
    {% blocktrans trimmed %}
      New users can be registered from any client if public registration is
      enabled.
    {% endblocktrans %}
  </p>

  {% if certificate_status != "valid" %}
    <div class="alert alert-warning" role="alert">
      {% url 'letsencrypt:index' as letsencrypt_url %}
      {% blocktrans trimmed %}
        The configured domain name is using a self-signed certificate.
        Federation with other Matrix Synapse instances requires a valid TLS
        certificate. Please go to <a href="{{ letsencrypt_url }}">Let's
        Encrypt</a> to obtain one.
      {% endblocktrans %}
    </div>
  {% endif %}
{% endblock %}

{% block page_js %}
  <script type="text/javascript"
          src="{% static 'matrixsynapse/matrixsynapse.js' %}"></script>
{% endblock %}
